import mock

from .support import unittest
from cliquet import get_session


class SessionCookiesTest(unittest.TestCase):

    def test_get_session_use_default_values_when_not_provided(self):
        pass
